/*
  Project Name: Initative
  Description: This program controls the display of the initiative. 
  An LED strip with 34 LEDs and a button are connected to the Arduino. 
  The ESP automatically connects to the server as soon as it is 
  started and then waits for messages. These messages can either set 
  the number of creatures and start the initiative or increase, 
  decrease, or delete the initiative. Once the initiative has started,
  it is represented by 2 LEDs and can be advanced by one position by
  pressing the button.
  Author: Lena Hagenauer
  Date: 05 January 2025
  Version: 2.0
  License: MIT License
*/

// Include libraries

// NeoPixel
#include <Adafruit_NeoPixel.h>
#ifdef __AVR__
#include <avr/power.h>
#endif
#define PIN        2
#define NUMPIXELS 34
Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

// Button
int pushButton = 4;

// Comunication Information
int nrCreaturs = 0; //number of how many playes
int stateInitaive = 99;
int currentCreatur = 1;

// WIFI & ESPNOW
#include <ESP8266WiFi.h>
#include <espnow.h>

uint8_t channel = 1;


#define MAX_CHANNEL 13  // 11 in North America or 13 in Europe

unsigned long currentMillis = millis();
unsigned long lastTime = 0;  
unsigned long timerDelay = 2000;  // send readings timer

uint8_t broadcastAddressX[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

uint8_t clientMacAddress[6];

enum PairingStatus {PAIR_REQUEST, PAIR_REQUESTED, PAIR_PAIRED, };

PairingStatus pairingStatus = PAIR_REQUEST;

enum MessageType {PAIRING, DATA, LIGHT, INITATIVE};
MessageType messageType;

// Updates DHT readings every 10 seconds
//const long interval = 10000;
unsigned long previousMillis = 0;    // will store last time DHT was updated


// Structure to send Informtion to Initative
typedef struct struct_initative {
  uint8_t msgType;
  int nr;
  int state; //add 1; sup 2; end 99;
} struct_initative;

// Structure for pairing message
typedef struct struct_pairing {      
  uint8_t msgType;
  uint8_t id;
  uint8_t macAddr[6];
  uint8_t channel;
  char name[64];
} struct_pairing;

// Variables definitions
struct_initative incomingReadings;

struct_pairing pairingData;

#define BOARD_ID 2
unsigned long start;
int state = 0;


// Callback when data is sent
void OnDataSent(uint8_t *mac_addr, uint8_t sendStatus) {
  Serial.print("Last Packet Send Status: ");
  if (sendStatus == 0){
    Serial.println("Delivery success");
  }
  else{
    Serial.println("Delivery fail");
  }
}

// Function to print MAC address
void printMAC(const uint8_t * mac_addr){
  char macStr[18];
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
  Serial.print(macStr);
}

// Callback when data is received
void OnDataRecv(uint8_t * mac, uint8_t *incomingData, uint8_t len) {
  Serial.print("Size of message : ");
  Serial.print(len);
  Serial.print(" from ");
  printMAC(mac);
  Serial.println();
  uint8_t type = incomingData[0];

  switch (type) {
  case INITATIVE:  
   
    memcpy(&incomingReadings, incomingData, sizeof(incomingReadings));
    Serial.print(len);
    Serial.print(" Data bytes received from: ");
    printMAC(mac);
    Serial.println();

    
    
    stateInitaive = incomingReadings.state;
    Serial.println(stateInitaive);
    if(stateInitaive == 0){
      nrCreaturs = incomingReadings.nr;
      Serial.println(nrCreaturs);
      currentCreatur = 1;
    }
    

    break;

  case PAIRING:
    memcpy(&pairingData, incomingData, sizeof(pairingData));
    if (pairingData.id == 0) {                // the message comes from server
      Serial.print("Pairing done for ");
      printMAC(pairingData.macAddr);
      Serial.print(" on channel " );
      Serial.print(pairingData.channel);    // channel used by the server
      Serial.print(" in ");
      Serial.print(millis()-start);
      Serial.println("ms");
      //esp_now_del_peer(pairingData.macAddr);
      //esp_now_del_peer(mac);
      esp_now_add_peer(pairingData.macAddr, ESP_NOW_ROLE_COMBO, pairingData.channel, NULL, 0); // add the server to the peer list 
      pairingStatus = PAIR_PAIRED ;            // set the pairing status
    }
    break;
  }  
}

// Function to read the MAC address
void readGetMacAddress(){
  String val = WiFi.macAddress();
  Serial.println(val);
  char* endPtr; 
  clientMacAddress[0] = strtol(val.c_str(), &endPtr, 16); // read the first starting at the beginning of the buffer. this initializes endPtr as a pointer to the ':' after the first number 
  for (int i = 1;  (*endPtr) && (i < 6); i++) {
    clientMacAddress[i] = strtol(endPtr + 1, &endPtr, 16); // using +1 for the pointer as we want to skip the ':'
  }

  for (int i = 0; i < 6; i++) {
    Serial.print(clientMacAddress[i], HEX);
    if (i != 5) Serial.print(F(":"));
  }
}

// Function for autoPairing
PairingStatus autoPairing(){
  switch(pairingStatus) {
  case PAIR_REQUEST:
    Serial.print("Pairing request on channel "  );
    Serial.println(channel);
  
    // clean esp now
    esp_now_deinit();
    WiFi.mode(WIFI_STA);
    // set WiFi channel   
    wifi_promiscuous_enable(1);
    wifi_set_channel(channel);
    wifi_promiscuous_enable(0);
    //WiFi.printDiag(Serial);
    WiFi.disconnect();

    // Init ESP-NOW
    if (esp_now_init() != 0) {
      Serial.println("Error initializing ESP-NOW");
    }
    esp_now_set_self_role(ESP_NOW_ROLE_COMBO);
    // set callback routines
    esp_now_register_send_cb(OnDataSent);
    esp_now_register_recv_cb(OnDataRecv);
    
    // set pairing data to send to the server
    pairingData.id = BOARD_ID;     
    pairingData.channel = channel;
    pairingData.macAddr[0] = clientMacAddress[0];
    pairingData.macAddr[1] = clientMacAddress[1];
    pairingData.macAddr[2] = clientMacAddress[2];
    pairingData.macAddr[3] = clientMacAddress[3];
    pairingData.macAddr[4] = clientMacAddress[4];
    pairingData.macAddr[5] = clientMacAddress[5];
    strcpy(pairingData.name, "");
 
    previousMillis = millis();
    // add peer and send request
    Serial.println(esp_now_send(broadcastAddressX, (uint8_t *) &pairingData, sizeof(pairingData)));
    pairingStatus = PAIR_REQUESTED;
    break;

  case PAIR_REQUESTED:
    // time out to allow receiving response from server
    currentMillis = millis();
    if(currentMillis - previousMillis > 1000) {
      previousMillis = currentMillis;
      // time out expired,  try next channel
      channel ++;
      if (channel > MAX_CHANNEL) {
        channel = 0;
      }
      pairingStatus = PAIR_REQUEST; 
    }
    break;

  case PAIR_PAIRED:
    //Serial.println("Paired!");
    break;
  }
  return pairingStatus;
} 




// Fungtioin to show rainbow
void rainbow(int wait) {
  for(long firstPixelHue = 0; firstPixelHue < 65536; firstPixelHue += 256) {

    pixels.rainbow(firstPixelHue,2);
    pixels.show();
    delay(wait);  
  }
}



void setup() {

  // Init Serial Monitor
  Serial.begin(74880);

  // Init neoPixel
  pixels.begin();

  // Init Button
  pinMode(pushButton, INPUT);

  // Set device as a Wi-Fi Station
  WiFi.mode(WIFI_STA);
  readGetMacAddress();

  // Serial.println(WiFi.macAddress());
  WiFi.disconnect();

  // Init ESP-NOW
  if (esp_now_init() != 0) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Set ESP-NOW Role
  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);
  
  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
  esp_now_register_send_cb(OnDataSent);

  pairingData.id = 1;

}

// Funktion to turn Led on 
void setCreatur(int nr){
  pixels.clear();
  nr = nr;
  pixels.setPixelColor((nr*2)-2, 255, 255, 255);
  pixels.setPixelColor((nr*2)-1, 255, 255, 255);
 
  pixels.show();
}

int togel = 0;
void loop() {

  if(state == 0){ // Wait for Pairing
    if(autoPairing() == PAIR_PAIRED ){
      state = 1;
      rainbow(10);
    }
  }else{ // Pairing was successful
    
    if(stateInitaive == 99){// end Initaive
      pixels.clear(); 
      pixels.show();
      currentCreatur = 1;
    }else if(stateInitaive == 1){// add player
      nrCreaturs++;
      stateInitaive = 0;
      if(nrCreaturs > 17){//To many players
        // Glow red for 1 second
        pixels.clear();
        for(int i = 0; i<35; i++){
          pixels.setPixelColor(i, 255, 0, 0);
          pixels.show();
        }

        delay(1000);
        pixels.clear(); 
        pixels.show();
        
        stateInitaive = 0;
      }
      
      
    }else if(stateInitaive == 2){// Delete player
      if(nrCreaturs == 1){// No players left stop initative
        stateInitaive = 99;
        nrCreaturs = 0;
      }else{
        nrCreaturs--;
        stateInitaive = 0;
      }
    }else{//start new initative
      setCreatur(currentCreatur);
      int buttonState = digitalRead(pushButton);
      
      if(buttonState == 1){
        togel = 1;
        
      }else{
        if(togel == 1){// Next player
        
          if(currentCreatur < nrCreaturs){
            currentCreatur++;           
          }else{
            currentCreatur = 1;
          }
          togel = 0;
        }
      }
    }
  }
}